type symbol = string
type var = string 

type descr =
| Link of  t
| Repr of constr

and constr =
| Var
| Op of symbol  * t list

and t = descr ref

type problem = (t * t) list * (var * t) list

(* Transforme une expression en une expression avec du partage.
 * La liste [defined : (var * t) list] contient la représentation
 * des variables déjà défines. La fonction retourne à la fois 
 * l'objet converti (c'est la première composante) et une mise 
 * à jour de la liste des variables (elle est mise à jour dans 
 * le cas où on rencontre pour la première fois une variable 
 * durant la conversion). 
 * *)
let rec sharing_of_expr defined = function
  | Expr.Var x -> 
      failwith "Sharing.sharing_of_expr: to be implemented"
  | Expr.Op (f, l) -> 
      failwith "Sharing.sharing_of_expr: to be implemented"

let rec repr t = 
  failwith "Sharing.repr: to be implemented"

exception Not_unifyable 

let rec unify t1 t2 = 
  let r1, r2 = repr t1, repr t2 in 
  match !r1, !r2 with 
  | Link _, _ | _, Link _ -> assert false
  | _ -> 
   failwith "Sharing.unify: to be implemented"

let cyclic t = 
  false (* To be implemented *)

(* Cette fonction ne boucle pas sur les termes infinis. *)
let string_of_sharing defined t = 
 let rec aux t = 
    try 
      List.assq t defined 
    with Not_found -> begin
      match !t with 
       | Link r -> aux r
       | Repr (Op (f, args)) -> Printf.sprintf "%s(%s)" f
           (String.concat ", " (List.map aux args))
       | Repr Var -> failwith "string_of_sharing"
    end
  in begin
      match !t with 
       | Link r -> aux r
       | Repr (Op (f, args)) -> Printf.sprintf "%s(%s)" f
           (String.concat ", " (List.map aux args))
       | Repr Var -> List.assq t defined 
    end


let string_of_sharing_swapped defined t = 
  string_of_sharing (List.map (fun (x,y) -> (y,x)) defined) t

let string_of_problem (eqns, defined) = 
    String.concat "\n" 
      (List.map (fun (x,y) -> 
        Printf.sprintf "%s = %s" 
         (string_of_sharing_swapped defined x)
         (string_of_sharing_swapped defined y)) eqns)
